/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2002 Ian Zimmerman <itz@speakeasy.org> */

static const char rcsid[] = "$Id: clientcon.c,v 1.18 2002/09/11 20:23:29 itz Exp $ ";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef _REENTRANT
#define _REENTRANT
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef MAJOR_IN_MKDEV
#include <sys/mkdev.h>
#endif 

#ifdef MAJOR_IN_SYSMACROS
#include <sys/sysmacros.h>
#endif

#include "readproc.h"
#include "clientcon.h"
#include "gim_errno.h"
#include "preadwrite.h"
#include "gim_event.h"
#include "exc.h"
#include "gim_common.h"

const char* gim_server_err_message;

void
gim_init ()
{
  exc_init();
}

static int
get_cterm ()
{
  struct proc_s myself;

  if (0 <= look_up_our_self(&myself) && myself.tty > 0) {
    return (major(myself.tty) == TTY_MAJOR ?  minor(myself.tty) : -1);
  } else {
    return -1;
  } /*if*/
}

static void
gim_send_masks (int fd, const struct gim_client_masks* p_masks)
{
  static const char minmod_format[] = "minmod %x\n";
  static const char maxmod_format[] = "maxmod %x\n";
  static const char handle_format[] = "handle %x\n";
  static const char pass_format[] = "pass %x\n";

  int len;
  char blurb[100];
  char server_response[512];
  char* p;
  int code;

  len = snprintf(blurb, sizeof(blurb), minmod_format, p_masks->minmod);
  assert(len >= 0 && len < sizeof(blurb));
  p = server_response;
  code = gim_transact(fd, blurb, &p, sizeof(server_response));
  if (code >= 400) {
    gim_server_err_message = p;
    exc_eraise(server_cmd_failure);
  } /*if*/
    
  len = snprintf(blurb, sizeof(blurb), maxmod_format, p_masks->maxmod);
  assert(len >= 0 && len < sizeof(blurb));
  p = server_response;
  code = gim_transact(fd, blurb, &p, sizeof(server_response));
  if (code >= 400) {
    gim_server_err_message = p;
    exc_eraise(server_cmd_failure);
  } /*if*/

  len = snprintf(blurb, sizeof(blurb), handle_format, p_masks->handle);
  assert(len >= 0 && len < sizeof(blurb));
  p = server_response;
  code = gim_transact(fd, blurb, &p, sizeof(server_response));
  if (code >= 400) {
    gim_server_err_message = p;
    exc_eraise(server_cmd_failure);
  } /*if*/
    
  len = snprintf(blurb, sizeof(blurb), pass_format, p_masks->pass);
  assert(len >= 0 && len < sizeof(blurb));
  p = server_response;
  code = gim_transact(fd, blurb, &p, sizeof(server_response));
  if (code >= 400) {
    gim_server_err_message = p;
    exc_eraise(server_cmd_failure);
  } /*if*/

}

int
gim_client_connect (const struct gim_client_masks* p_masks, int vt, const char* device)
{
  struct sockaddr_un server;
  char blurb[100];
  char server_response[512];
  char server_protocol[10];
  size_t len;
  int i;
  int items;
  int code;
  int status;
  int fd;
  char* p;
  int server_prot_current, server_prot_revision, server_prot_age;
  int client_prot_current, client_prot_revision, client_prot_age;

  static const char open_format[] = "open %d\n";
  static const char protocol_format[] = "%d:%d:%d";

  /* if caller didn't pass a vt#, try to guess the controlling terminal */
  if (vt < 0) vt = get_cterm();

  /* if that didn't help, we give up. */
  if (vt < 0) return GIM_CONNECT_NO_CTTY;

  fd = -1;
  if (device == 0) device = "mouse";

  exc_try {
    if ((fd = socket(AF_LOCAL, SOCK_STREAM, 0)) == -1) exc_eraise2(unix_error, errno); 
    server.sun_family = AF_LOCAL;
    if (0 > (i = snprintf(server.sun_path, sizeof(server.sun_path), gim_sock_path, device)) ||
        i >= sizeof(server.sun_path))
      exc_eraise(buffer_too_small);
    if (0 > connect(fd, (struct sockaddr*)&server, sizeof(struct sockaddr_un)))
      exc_eraise2(unix_error, errno);
    len = gim_pread(fd, blurb, sizeof(blurb));
    if (len == 0 || blurb[len-1] != '\n') exc_eraise(invalid_greeting);

    /* grab and verify protocol version */
    items = sscanf(blurb, gim_greeting_format, server_protocol);
    if (items < 1) exc_eraise(invalid_greeting);
    items = sscanf(server_protocol, protocol_format,
                   &server_prot_current, &server_prot_revision, &server_prot_age);
    if (items < 3) exc_eraise(invalid_greeting);
    items = sscanf(PROTOCOL, protocol_format,
                   &client_prot_current, &client_prot_revision, &client_prot_age);
    assert(items == 3);
    if (client_prot_current - client_prot_age < server_prot_current - server_prot_age ||
        client_prot_current > server_prot_current) 
      exc_eraise(version_mismatch);

    /* send open command to server */
    len = snprintf(blurb, sizeof(blurb), open_format, vt);
    assert(len >= 0 && len < sizeof(blurb));
    p = server_response;
    code = gim_transact(fd, blurb, &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    /* open succeeded, set the parameters */
    gim_send_masks(fd, p_masks);

    /* ready to go - tell server to start! */
    p = server_response;
    code = gim_transact(fd, "start\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    status = fd;
  }
  exc_except(unix_error) {
    status = GIM_CONNECT_ERR_UNIX;
  }
  exc_except(invalid_greeting) {
    status = GIM_CONNECT_ERR_GREETING;
  }
  exc_except(server_cmd_failure) {
    status = GIM_CONNECT_CMD_FAILURE;
  }
  exc_except(version_mismatch) {
    status = GIM_CONNECT_ERR_VERSION; 
  } 
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    status = GIM_CONNECT_ERR_UNKNOWN;
  } exc_end;

  if (status < 0 && fd != -1) close(fd);
  return status;
}

int
gim_client_close (int fd)
{
  char* p;
  int status;
  int code;
  char server_response[512];

  p = server_response;
  exc_try {
    code = gim_transact(fd, "close\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/
    close(fd);
    status = GIM_CLOSE_SUCCESS;
  }
  exc_except(server_cmd_failure) {
    status = GIM_CLOSE_CMD_FAILURE;
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    status = GIM_CLOSE_ERR_UNKNOWN;
  } exc_end;

  return status;
}

int
gim_client_set_masks (int fd, const struct gim_client_masks* p_masks)
{
  char* p;
  int code;
  char blurb[100];
  char server_response[512];

  p = server_response;

  exc_try {
    code = gim_transact(fd, "stop\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    /* stop succeeded, set the parameters */
    gim_send_masks(fd, p_masks);

    /* ready to go - tell server to start! */
    p = server_response;
    code = gim_transact(fd, "start\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    exc_ereturn2(int, GIM_SET_MASKS_SUCCESS); 
  }
  exc_except(server_cmd_failure) {
    exc_ereturn2(int, GIM_SET_MASKS_CMD_FAILURE);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_SET_MASKS_ERR_UNKNOWN);
  } exc_end;

}

int
gim_client_get_masks (int fd, struct gim_client_masks* p_masks)
{
  char* p;
  int code;
  char server_response[512];

  p = server_response;

  exc_try {
    code = gim_transact(fd, "getmasks\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    if (4 != sscanf(p, " min %x max %x handle %x pass %x",
                    p_masks->minmod, p_masks->maxmod,
                    p_masks->handle, p_masks->pass)) {
      exc_eraise(protocol_error);
    } /*if*/

    exc_ereturn2(int, GIM_GET_MASKS_SUCCESS);
  }
  exc_except(server_cmd_failure) {
    exc_ereturn2(int, GIM_GET_MASKS_CMD_FAILURE);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_GET_MASKS_ERR_UNKNOWN);
  } exc_end;

}

void
gim_get_event_internal (int fd, struct gim_event_t* p_ev)
{
  int code;
  int items;
  char server_response[512];
  char *p;

  gim_pread(fd, server_response, sizeof(server_response));
  code = strtol(server_response, &p, 10);
  switch(code) {
  case 110:
    items = sscanf(p, " %d*%x {%d,%d,%d} %dx%d %u.%u [%x.%x]",
                   &p_ev->vt, &p_ev->kind, &p_ev->x, &p_ev->y, &p_ev->z,
                   &p_ev->u.b.repeat, &p_ev->u.b.which,
                   &p_ev->time.tv_sec, &p_ev->time.tv_usec,
                   &p_ev->state, &p_ev->modifiers);
    if (items != 11) exc_eraise(protocol_error);
    break;
  case 111:
    items = sscanf(p, " %d*%x {%d(%d),%d(%d),%d(%d)} |%x %u.%u [%x.%x]",
                   &p_ev->vt, &p_ev->kind,
                   &p_ev->x, &p_ev->u.d.x, &p_ev->y, &p_ev->u.d.y, &p_ev->z, &p_ev->u.d.z, 
                   &p_ev->u.d.margin, &p_ev->time.tv_sec, &p_ev->time.tv_usec,
                   &p_ev->state, &p_ev->modifiers);
    if (items != 13) exc_eraise(protocol_error);
    break;
  default:
    exc_eraise(protocol_error);
  } /*switch*/

}

int
gim_get_event (int fd, struct gim_event_t* p_ev)
{
  int code;
  int items;
  char server_response[512];
  char *p;

  exc_try {
    gim_get_event_internal(fd, p_ev);
    exc_ereturn2(int, GIM_GET_EVENT_SUCCESS); 
  }
  exc_except(protocol_error) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_GET_EVENT_ERR_PROTOCOL);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_GET_EVENT_ERR_UNKNOWN);
  } exc_end;

}

int
gim_client_draw_pointer (int fd, int x, int y)
{
  char *p;
  int code;
  char server_response[512];
  int len;
  char blurb[100];
  static const char drawptr_format[] = "drawptr %d %d\n";

  p = server_response;
  exc_try {
    code = gim_transact(fd, "stop\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    len = snprintf(blurb, sizeof(blurb), drawptr_format, x, y);
    assert (len > 0 || len < sizeof(blurb));
    
    code = gim_transact(fd, blurb, &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/
    
    p = server_response;
    code = gim_transact(fd, "start\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    exc_ereturn2(int, GIM_DRAW_POINTER_SUCCESS);
  }
  exc_except(server_cmd_failure) {
    exc_ereturn2(int, GIM_DRAW_POINTER_CMD_FAILURE);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_DRAW_POINTER_ERR_UNKNOWN);
  } exc_end;

}

int
gim_client_stop (int fd)
{
  char *p;
  int code;
  char server_response[512];

  p = server_response;
  exc_try {
    code = gim_transact(fd, "stop\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    exc_ereturn2(int, GIM_STOP_SUCCESS);
  }
  exc_except(server_cmd_failure) {
    exc_ereturn2(int, GIM_STOP_CMD_FAILURE);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_STOP_ERR_UNKNOWN);
  } exc_end;

}

int
gim_client_start (int fd)
{
  char *p;
  int code;
  char server_response[512];

  p = server_response;
  exc_try {
    code = gim_transact(fd, "start\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    exc_ereturn2(int, GIM_START_SUCCESS);
  }
  exc_except(server_cmd_failure) {
    exc_ereturn2(int, GIM_START_CMD_FAILURE);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_START_ERR_UNKNOWN);
  } exc_end;

}

int
gim_client_devinfo (int fd, char* info, size_t info_size)
{
  char *p;
  int code;
  char server_response[512];
  int len;
  
  p = server_response;
  exc_try {
    code = gim_transact(fd, "devinfo\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    len = strlen(p);
    if (len >= info_size) {
      exc_eraise2(buffer_too_small, len);
    }

    strncpy(info, p, info_size);
    exc_ereturn2(int, len);
  }
  exc_except(server_cmd_failure) {
    exc_ereturn2(int, GIM_DEVINFO_CMD_FAILURE);
  }
  exc_except(buffer_too_small) {
    exc_ereturn2(int, *(int *)exc_info);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_DEVINFO_ERR_UNKNOWN);
  } exc_end;
}

int
gim_client_servinfo (int fd, char* info, size_t info_size)
{
  char *p;
  int code;
  char server_response[512];
  int len;
  
  p = server_response;
  exc_try {
    code = gim_transact(fd, "servinfo\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    len = strlen(p);
    if (len >= info_size) {
      exc_eraise2(buffer_too_small, len);
    }

    strncpy(info, p, info_size);
    exc_ereturn2(int, len);
  }
  exc_except(server_cmd_failure) {
    exc_ereturn2(int, GIM_SERVINFO_CMD_FAILURE);
  }
  exc_except(buffer_too_small) {
    exc_ereturn2(int, *(int *)exc_info);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_SERVINFO_ERR_UNKNOWN);
  } exc_end;
}

int
gim_client_disable_paste (int fd)
{
  char *p;
  int code;
  char server_response[512];
  int len;
  
  p = server_response;
  exc_try {
    code = gim_transact(fd, "disable_paste\n", &p, sizeof(server_response));
    if (code >= 400) {
      gim_server_err_message = p;
      exc_eraise(server_cmd_failure);
    } /*if*/

    exc_ereturn2(int, GIM_DISABLE_PASTE_SUCCESS);
  }
  exc_except(server_cmd_failure) {
    exc_ereturn2(int, GIM_DISABLE_PASTE_CMD_FAILURE);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, GIM_DISABLE_PASTE_ERR_UNKNOWN);
  } exc_end;
}
